from ._tk_vars import StringVar, IntVar, DoubleVar, BooleanVar, Variable
from ._notifications import KeyboardNotifyEvent
from ._geo_types import Layout, GeoNotes
from ._better_dict import BetterDict
from ._binds import FrameBind
from ._scheme import Scheme
from ._style import Style
from ._constants import *
from ._color import Color



"""
_constants.py
05. February 2023

really don't know why I put this file in types, but here are a few constants

Author:
Nilusink
"""
# directional
LEFT: str = "left"
RIGHT: str = "right"
TOP: str = "top"
BOTTOM: str = "bottom"

# sticky (no I didn't put in all combinations because you
# can just make them yourself, the order doesn't matter)
N: str = "n"
S: str = "s"
E: str = "e"
W: str = "e"
NS: str = N + S
SW: str = S + W
NSEW: str = N + S + E + W



"""
_pg_root.py
04. February 2023

the root of the window

Author:
Nilusink
"""
from concurrent.futures import ThreadPoolExecutor as Pool, Future
from .widgets import GeometryManager
from .theme import ThemeManager
from time import perf_counter
from copy import deepcopy
from .types import *
import typing as tp
import pygame as pg
import os.path


DEFAULT_TITLE: str = "Window"
DEFAULT_ICON: str = os.path.dirname(__file__) + "/icon.png"


RES_T = tp.TypeVar("RES_T")


class _TimeoutCandidate(tp.TypedDict):
    timeout_left: float
    function: tp.Callable[[tp.Any], RES_T]
    future: Future[RES_T]
    args: tuple[tp.Any, ...]
    kwargs: dict[str, tp.Any]


class PgRoot(GeometryManager):
    _focus_item: GeometryManager | None = None
    _running: bool = True
    _theme: ThemeManager = ...
    __background: pg.Surface = ...
    layout_params: BetterDict = ...
    _min_size: tuple[int, int] = ...
    _bg_configured: bool = False
    _mouse_pos: tuple[int, int] = ...
    _max_framerate: int = ...

    __timeouts: list[_TimeoutCandidate]
    _last_it_call: float
    _tpool: Pool

    def __init__(
            self,
            title: str = ...,
            icon_path: str = ...,
            size: tuple[int, int] = ...,
            bg_color: Color = ...,
            padding: int = 0,
            margin: int = 0,
            max_framerate: int = 30
    ):
        self._last_it_call = perf_counter()
        self._max_framerate = max_framerate
        self._tpool = Pool(max_workers=10)
        self.__timeouts = []

        super().__init__()
        self._theme = ThemeManager()
        self._theme.notify_on(ThemeManager.NotifyEvent.theme_reload, self.notify)
        self._min_size = (0, 0)

        # args
        self._bg_configured = bg_color is not ...
        self._bg = self._theme.root.bg.hex if bg_color is ... else bg_color

        self._layout_params = BetterDict({
            "padding": padding,
            "margin": margin,
        })

        # pg init
        pg.init()
        pg.font.init()
        self._clk = pg.time.Clock()

        if size is not ...:
            self.__background = pg.display.set_mode(size, flags=pg.RESIZABLE)

        else:
            self.__background = pg.display.set_mode(flags=pg.RESIZABLE)

        # set icon and caption
        pg.display.set_caption(DEFAULT_TITLE if title is ... else title)
        img = pg.image.load(
            DEFAULT_ICON if icon_path is ... else icon_path, "icon"
        )
        pg.display.set_icon(img)

    # config
    @property
    def title(self) -> str:
        return pg.display.get_caption()[0]

    @title.setter
    def title(self, value: str) -> None:
        pg.display.set_caption(value)

    @property
    def theme(self) -> ThemeManager:
        return self._theme

    @property
    def mouse_pos(self) -> tuple[int, int]:
        return self._mouse_pos

    @property
    def root(self) -> tp.Self:
        return self

    def get_focus(self) -> GeometryManager | None:
        """
        get the currently focused item
        """
        return self._focus_item

    def set_focus(
            self,
            widget: tp.Union["GeometryManager", None, tp.Any] = None
    ) -> None:
        """
        set the focused item
        """
        if self._focus_item is not None:
            self._focus_item.stop_focus()

        if widget is not None:
            widget.set_focus()
            self._focus_item = widget

    def notify_focus(self, widget: tp.Union["GeometryManager", None] = None):
        """
        notify the root that a widget has been set as focus
        """
        if self._focus_item != widget:
            self.set_focus(widget)

    # interfacing
    def notify(self, event: ThemeManager.NotifyEvent, _info=...) -> None:
        """
        gets called by another class
        """
        match event:
            case ThemeManager.NotifyEvent.theme_reload:
                # the theme has been reloaded
                if not self._bg_configured:
                    self._bg = self.theme.root.bg.rgba

    # pygame stuff
    def _event_handler(self) -> None:
        """
        handle the events raised by pygame
        """
        for event in pg.event.get():
            match event.type:
                case pg.QUIT:
                    self._running = False

                case pg.KEYDOWN:
                    if self._focus_item is not None:
                        self._focus_item.notify(
                            KeyboardNotifyEvent.key_down,
                            event
                        )

                case pg.KEYUP:
                    if self._focus_item is not None:
                        self._focus_item.notify(
                            KeyboardNotifyEvent.key_up,
                            event
                        )

                # case pg.VIDEORESIZE:  # window size changed
                #     width, height = event.size
                #
                #     print("updating size: ", (width, height), "\t", self._min_size)
                #
                #     width = max([width, self._min_size[0]])
                #     height = max([height, self._min_size[1]])
                #
                #     # self.__background = pg.display.set_mode((width, height), flags=pg.RESIZABLE | pg.HWSURFACE | pg.DOUBLEBUF)

        self._mouse_pos = pg.mouse.get_pos()

        self._notify_child_active_hover(self.mouse_pos)

    def update_idletasks(self) -> None:
        """
        updates all the functional tasks
        """
        self._event_handler()

        # handle timeouts
        for to in self.__timeouts.copy():
            # calculate time since last function call
            now = perf_counter()
            delta = now - self._last_it_call
            self._last_it_call = now

            # remove time from timeout
            to["timeout_left"] -= delta * 1000

            # if the timeout is finished, execute the function
            if to["timeout_left"] <= 0:
                # create a function that populates the timeout future's result
                # with the function return value
                def curr_func():
                    to["future"].set_result(to["function"](
                        *to["args"], **to["kwargs"]
                    ))

                # submit the created function to the threadpool
                # and remove it from the active timeouts
                self._tpool.submit(curr_func)
                self.__timeouts.remove(to)

    def update(self) -> None:
        """
        update the screen
        """
        self.__background.fill(self._bg)

        self.calculate_geometry()
        for child, params in self._child_params:
            child.draw(self.__background)

        pg.display.flip()

    def mainloop(self):
        """
        run the windows main loop
        """
        while self._running:
            self.update_idletasks()
            self.update()

            self._clk.tick(self._max_framerate)

    def calculate_geometry(self):
        """
        calculate how each individual child should be placed
        """
        match self._layout:
            case Layout.Absolute:  # Absolute
                # since the positioning is absolute, the children should not influence the parents size
                for child, params in self._child_params:
                    child.set_position(params.x, params.y)

                return

            case Layout.Pack:
                directional_dict: dict[str, int | list] = {"total_x": 0, "total_y": 0, "children": [], "sizes": []}
                top = deepcopy(directional_dict)
                bottom = deepcopy(directional_dict)
                left = deepcopy(directional_dict)
                right = deepcopy(directional_dict)

                # get all sizes and group by anchor
                for child, param in self._child_params:
                    child_size = child.calculate_size()

                    if param.anchor == TOP:
                        top["children"].append(child)
                        top["sizes"].append(child_size)
                        top["total_x"] += child_size[0]
                        top["total_y"] += child_size[1]

                    elif param.anchor == BOTTOM:
                        bottom["children"].append(child)
                        bottom["sizes"].append(child_size)
                        bottom["total_x"] += child_size[0]
                        bottom["total_y"] += child_size[1]

                    elif param.anchor == LEFT:
                        left["children"].append(child)
                        left["sizes"].append(child_size)
                        left["total_x"] += child_size[0]
                        left["total_y"] += child_size[1]

                    elif param.anchor == RIGHT:
                        right["children"].append(child)
                        right["sizes"].append(child_size)
                        right["total_x"] += child_size[0]
                        right["total_y"] += child_size[1]

                top["total_y"] += self._layout_params.padding * len(top["children"]) - 1
                bottom["total_y"] += self._layout_params.padding * len(bottom["children"]) - 1

                left["total_x"] += self._layout_params.padding * len(left["children"]) - 1
                right["total_x"] += self._layout_params.padding * len(right["children"]) - 1

                min_x = max([top["total_x"], bottom["total_x"], left["total_x"] + right["total_x"]])
                min_y = max([left["total_y"], right["total_y"], top["total_y"] + bottom["total_y"]])

                # add margin
                min_x += self.layout_params.margin * 2
                min_y += self.layout_params.margin * 2

                self._min_size = min_x, min_y

                total_x, total_y = self.calculate_size()

                # tell the children where they should be
                y_cen = total_y / 2
                x_cen = total_x / 2

                # left
                x_now = self._layout_params.margin
                for child, size in zip(left["children"], left["sizes"]):
                    child.set_position(x_now, y_cen - size[1] / 2)
                    x_now += size[0] + self._layout_params.padding

                    # right
                x_now = total_x - self._layout_params.margin
                for child, size in zip(right["children"], right["sizes"]):
                    child.set_position(x_now - size[0], y_cen - size[1] / 2)
                    x_now -= size[0] + self._layout_params.padding

                    # top
                y_now = self._layout_params.margin
                for child, size in zip(top["children"], top["sizes"]):
                    child.set_position(x_cen - size[0] / 2, y_now)
                    y_now += size[1] + self._layout_params.padding

                    # bottom
                y_now = total_y - self._layout_params.margin
                for child, size in zip(bottom["children"], bottom["sizes"]):
                    child.set_position(x_cen - size[0] / 2, y_now - size[1])
                    y_now -= size[1] + self._layout_params.padding

            case Layout.Grid:
                rows: list[dict[str, tp.Any | float]] = []
                columns: list[dict[str, tp.Any | float]] = []

                for child, params in self._child_params:
                    row, column = params["row"], params["column"]

                    # if row was not yet made, make all previous ones
                    if len(rows) <= row:
                        for n_row in range(len(rows), row + 1):
                            out = {
                                "weight": 0,
                                "children": []
                            }

                            if n_row in self._grid_params.rows:
                                config = self._grid_params.rows[n_row]

                                if "weight" in config:
                                    out["weight"] = config["weight"]

                            rows.append(out)

                    if len(columns) <= column:
                        for n_col in range(len(columns), column + 1):
                            out = {
                                "weight": 0,
                                "children": []
                            }

                            if n_col in self._grid_params.columns:
                                config = self._grid_params.columns[n_col]

                                if "weight" in config:
                                    out["weight"] = config["weight"]

                            columns.append(out)

                    rows[row]["children"].append((child, params))
                    columns[column]["children"].append((child, params))

                matrix: list[list] = []

                for r in range(len(rows)):
                    matrix.append([])

                    for c in range(len(columns)):
                        # child = set(rows[r]["children"]) & set(columns[c]["children"])
                        child = [chi for chi in rows[r]["children"] if chi in columns[c]["children"]]
                        child = list(child)

                        if len(child) > 1:
                            raise ValueError(f"{len(child)} children assigned to row {r} column {c}!")

                        if child:
                            matrix[r].append(child[0])

                        else:
                            matrix[r].append(...)

                # calculate the minimal size for each row
                for r, row in enumerate(rows):
                    rows[r]["max_size"] = 0

                    for child, params in row["children"]:
                        _, y = child.calculate_size()

                        y += 2 * params.margin

                        if y > rows[r]["max_size"]:
                            rows[r]["max_size"] = y

                # calculate the minimal size for each column
                for c, column in enumerate(columns):
                    columns[c]["max_size"] = 0

                    for child, params in column["children"]:
                        x, _ = child.calculate_size()

                        x += 2 * params.margin

                        if x > columns[c]["max_size"]:
                            columns[c]["max_size"] = x

                    # calculate the container size
                width, height = self.calculate_size()

                # only subtract rows that don't have a weight
                min_width = sum([c["max_size"] for c in columns])
                min_height = sum([r["max_size"] for r in rows])

                # set the windows minimal size
                self._min_size = min_width, min_height

                # assign extra space
                extra_width = width - sum([c["max_size"] for c in columns if c["weight"] == 0])
                extra_height = height - sum([r["max_size"] for r in rows if r["weight"] == 0])

                total_row_weight = sum([row["weight"] for row in rows])
                total_column_weight = sum([column["weight"] for column in columns])

                # print(f"\n\nwindow_size=[{width}, {height}]\tmin_size={[min_width, min_height]}")
                # print(f"{total_row_weight=}")
                # print(f"{total_column_weight=}")
                # print(f"{extra_height=}")
                # print(f"{extra_width=}")

                # assign each row and column a specific size
                for r in range(len(rows)):
                    if total_row_weight == 0:
                        rows[r]["height"] = 0
                    else:
                        # assign either the minimum size or the calculated dynamic one
                        w_size = ((rows[r]["weight"] / total_row_weight) * extra_height).__floor__()
                        rows[r]["height"] = max([w_size, rows[r]["max_size"]])
                        # print(f"{w_size=}\t{rows[r]['max_size']=}")

                    # rows[r]["height"] += rows[r]["max_size"]
                    rows[r]["y_start"] = sum([prev_row["height"] for prev_row in rows[:r]])

                    for c in range(len(columns)):
                        if total_column_weight == 0:
                            columns[c]["width"] = 0
                        else:
                            # assign either the minimum size or the calculated dynamic one
                            w_size = ((columns[c]["weight"] / total_column_weight) * extra_width).__floor__()
                            columns[c]["width"] = max([w_size, columns[c]["max_size"]])
                            # print(f"{w_size=}\t{columns[c]['max_size']=}")

                        # columns[c]["width"] += columns[c]["max_size"]
                        columns[c]["x_start"] = sum([prev_col["width"] for prev_col in columns[:c]])

                        # pg.draw.rect(
                        #     self.__background,
                        #     (255, 0, 0, 255),
                        #     pg.Rect(columns[c]["x_start"], rows[r]["y_start"], columns[c]["width"], rows[r]["height"]),
                        #     width=1
                        # )

                # place children
                for child, params in self._child_params:
                    # place the child proportional to the table and stickiness
                    # size = list(child.calculate_size())
                    size = [child._width, child._height]

                    row, column = params["row"], params["column"]
                    sticky = params["sticky"]

                    width = columns[column]["width"]
                    height = rows[row]["height"]

                    x = columns[column]["x_start"]
                    y = rows[row]["y_start"]

                    x_cen = x + width / 2
                    y_cen = y + height / 2

                    x_diff = width - size[0]
                    y_diff = height - size[1]

                    # print(f"calc: {x_diff}, {y_diff}\t{size}\t{width},{height}")
                    # print(f"{sticky=}")

                    box_x = x_cen - size[0] / 2
                    box_y = y_cen - size[1] / 2

                    # assign stickiness
                    if not child._width_configured:
                        if "w" in sticky:
                            size[0] += (x_diff / 2) - params.margin
                            box_x = x + params.margin

                        if "e" in sticky:
                            size[0] += (x_diff / 2) - params.margin

                        child.assigned_width = size[0]

                    if not child._height_configured:
                        if "n" in sticky:
                            size[1] += (y_diff / 2) - params.margin
                            box_y = y + params.margin
                            # print("north: ", size, box_x, box_y, "\t\t", width, height)

                        if "s" in sticky:
                            size[1] += (y_diff / 2) - params.margin
                            # print("south: ", size, box_x, box_y, "\t\t", width, height)

                        child.assigned_height = size[1]

                    child.set_position(box_x, box_y)

            case _:
                raise ValueError(f"Invalid geometry type: {self._layout.__class__.__name__}")

    def calculate_size(self) -> tuple[int, int]:
        """
        calculate how big the container should be
        """
        # make sure the geometry is up-to-date
        return pg.display.get_window_size()

    # tool functions
    def after(
            self,
            timeout: int,
            function: tp.Callable[[tp.Any], RES_T],
            *args,
            **kwargs
    ) -> Future[RES_T]:
        """
        calls the given function after timeout milliseconds

        :param timeout: timeout in milliseconds
        :param function: the function to call
        :param args: the given functions positional arguments
        :param kwargs: the given functions keyword arguments
        :returns: a future with the function's result
        """
        n_future = Future[RES_T]()

        self.__timeouts.append({
            "timeout_left": timeout,
            "function": function,
            "future": n_future,
            "args": args,
            "kwargs": kwargs
        })

        return n_future



"""
_style.py
11. May 2023

The style of one element

Author:
Nilusink
"""
from ._geo_types import Layout
from ._color import Color
from enum import Enum
import typing as tp


class _NotifyEvent(Enum):
    property_change = 0


class Style:
    """
    What would a UI be without styles? Correct, a terminal!
    """
    NotifyEvent = _NotifyEvent

    width: int = ...
    height: int = ...
    minWidth: int = ...
    minHeight: int = ...

    color: Color = ...
    backgroundColor: Color = ...

    layout: Layout = ...
    margin: int = ...
    padding: int = ...

    borderRadius: int = ...
    borderBottomRadius: int = ...
    borderTopRadius: int = ...
    borderLeftRadius: int = ...
    borderRightRadius: int = ...
    borderBottomLeftRadius: int = ...
    borderBottomRightRadius: int = ...
    borderTopLeftRadius: int = ...
    borderTopRightRadius: int = ...
    borderWidth: int = ...
    borderColor: Color = ...

    fontSize: int = ...

    __notifiers: list[tuple[str, tp.Callable[[NotifyEvent, str], tp.Any]]]

    def __init__(
            self,
            width: int = ...,
            height: int = ...,
            minWidth: int = ...,
            minHeight: int = ...,
            color: Color = ...,
            backgroundColor: Color = ...,
            layout: Layout = ...,
            margin: int = ...,
            padding: int = ...,
            borderRadius: int = ...,
            borderBottomRadius: int = ...,
            borderTopRadius: int = ...,
            borderLeftRadius: int = ...,
            borderRightRadius: int = ...,
            borderBottomLeftRadius: int = ...,
            borderBottomRightRadius: int = ...,
            borderTopLeftRadius: int = ...,
            borderTopRightRadius: int = ...,
            borderWidth: int = ...,
            borderColor: Color = ...,
            fontSize: int = ...,
    ) -> None:
        """
        create a style element
        """
        self.__notifiers = []

        self.width = width
        self.height = height
        self.minWidth = minWidth
        self.minHeight = minHeight
        self.color = color
        self.backgroundColor = backgroundColor
        self.layout = layout
        self.margin = margin
        self.padding = padding
        self.borderRadius = borderRadius
        self.borderBottomRadius = borderBottomRadius
        self.borderTopRadius = borderTopRadius
        self.borderLeftRadius = borderLeftRadius
        self.borderRightRadius = borderRightRadius
        self.borderBottomLeftRadius = borderBottomLeftRadius
        self.borderBottomRightRadius = borderBottomRightRadius
        self.borderTopLeftRadius = borderTopLeftRadius
        self.borderTopRightRadius = borderTopRightRadius
        self.borderWidth = borderWidth
        self.borderColor = borderColor
        self.fontSize = fontSize

    @classmethod
    def from_dict(cls, ignore_invalid: bool = False, **properties) -> tp.Self:
        """
        Create a style element from a dict.
        """
        new_instance = cls()

        for prop, value in properties.items():
            try:
                new_instance[prop] = value

            except KeyError:
                if ignore_invalid:
                    continue

                raise

        return new_instance

    @property
    def properties(self) -> list[str]:
        """
        all available style properties
        """
        out = []
        for prop in self.__dict__:
            if prop.startswith("_") or prop.endswith("_"):
                continue

            if hasattr(self, prop):
                out.append(prop)

        return out

    def overwrite(self, other: tp.Self) -> tp.Self:
        """
        "merge" two styles, choosing the "other" styles values
        for doubles
        """
        new_instance = self.__class__()
        for prop in self.properties:
            # if available, use "other" style
            if other[prop] is not ...:
                new_instance[prop] = other[prop]

            # if "other" isn't available, try own style
            elif self[prop] is not ...:
                new_instance[prop] = self[prop]

        return new_instance

    def notify_on(
            self,
            variable: str,
            callback: tp.Callable[[NotifyEvent, str], tp.Any]
    ) -> None:
        """
        notify on variable change
        """
        self.__notifiers.append((variable, callback))

    # accessibility
    def __getitem__(self, item: str) -> tp.Any:
        if item in self.properties:
            return self.__dict__[item]

        raise KeyError(f"Can't find item \"{item}\" in Style.")

    def __setitem__(self, key: str, value: tp.Any) -> None:
        if key in self.properties:
            return self.__setattr__(key, value)

        raise KeyError(f"Can't find item \"{key}\" in Style.")

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def __setattr__(self, key: str, value: tp.Any) -> None:
        # border shorthands
        if key == "borderRadius":
            self.__setattr__("borderTopLeftRadius", value)
            self.__setattr__("borderTopRightRadius", value)
            self.__setattr__("borderBottomLeftRadius", value)
            self.__setattr__("borderBottomRightRadius", value)
            return

        elif key == "borderTopRadius":
            self.__setattr__("borderTopLeftRadius", value)
            self.__setattr__("borderTopRightRadius", value)
            return

        elif key == "borderBottomRadius":
            self.__setattr__("borderBottomLeftRadius", value)
            self.__setattr__("borderBottomRightRadius", value)
            return

        elif key == "borderLeftRadius":
            self.__setattr__("borderTopLeftRadius", value)
            self.__setattr__("borderBottomLeftRadius", value)
            return

        elif key == "borderRightRadius":
            self.__setattr__("borderTopRightRadius", value)
            self.__setattr__("borderBottomRightRadius", value)
            return

        if not key.endswith("__notifiers"):
            for n_key, callback in self.__notifiers:
                if n_key.lower() == key.lower():
                    callback(_NotifyEvent.property_change, n_key)

        # change regardless of notification
        self.__dict__[key] = value

    def __repr__(self) -> str:
        out = f"{super().__repr__()}: ""{\n"
        for prop in self.properties:
            out += f"\t{prop}: {self[prop]},\n"

        return out + "}\n"



"""
_frame.py
04. February 2023

Frame - the base widget

Author:
Nilusink
"""
from ._geo_manager import GeometryManager
from concurrent.futures import Future
from ..theme import ThemeManager
from ..types import *
import typing as tp
import pygame as pg

# try importing pil. The user shouldn't be forced to install pil, but
# will get an error if they try to use the image widget without having it
# installed
try:
    # noinspection PyUnresolvedReferences
    from PIL import Image
    PIL_EXISTS = True

except ImportError:
    PIL_EXISTS = False


if tp.TYPE_CHECKING:
    from .._pg_root import PgRoot


RES_T = tp.TypeVar("RES_T")


def display_configurify(key: str) -> str:
    """
    convert a Frame init key to it's corresponding DisplayConfig key
    """
    replaces = [
        ("bg_color", "bg"),
        ("border_bottom_left_radius", "blr"),
        ("border_bottom_right_radius", "brr"),
        ("border_top_left_radius", "ulr"),
        ("border_top_right_radius", "urr"),
    ]

    for init, config in replaces:
        if key == init:
            return config

    return key


def pi_image_to_surface(pil_image):
    """
    helper function for converting pillow images to pygame images
    """
    return pg.image.fromstring(
        pil_image.tobytes(),
        pil_image.size,
        pil_image.mode
    ).convert()


class _Bind(tp.TypedDict):
    id: int
    event: FrameBind
    function: tp.Callable[[tp.Any], tp.Any | None]


class Frame(GeometryManager):
    """
    The base widget
    """
    __parent: tp.Union["Frame", tp.Any] = ...
    _focused: bool = False
    _binds: list[_Bind]
    style: Style = ...
    hover_style: Style = ...
    active_style: Style = ...
    _x: int = -1
    _y: int = -1

    _image: Image.Image

    def __init__(
            self,
            parent: tp.Union["Frame", tp.Any],
            width: int = ...,
            height: int = ...,
            bg: Color = ...,
            layout: Layout = Layout.Absolute,
            border_radius: int = ...,
            border_bottom_radius: int = ...,
            border_top_radius: int = ...,
            border_bottom_left_radius: int = ...,
            border_bottom_right_radius: int = ...,
            border_top_left_radius: int = ...,
            border_top_right_radius: int = ...,
            border_width: int = ...,
            border_color: Color = ...,
            image: Image.Image = ...,
            margin: int = ...,
            padding: int = ...,
            min_width: int = ...,
            min_height: int = ...,
            style: Style = ...,
            active_style: Style = ...,
            hover_style: Style = ...
    ) -> None:
        """
        The most basic widget: a frame.
        When using the  style  property, all other styles will be overwritten!

        :param parent: the frames parent container
        :param width: width of the frame
        :param height: height of the frame
        :param bg: background color of the frame
        :param border_radius: border radius of the box (all four corners)
        :param border_width: how thick the border of the box should be
        :param border_color: the color of the border
        """
        if image is not ... and not PIL_EXISTS:
            raise ImportError(
                "PIL needs to be installed to use the Image widget.\n"
                "\tto install PIL, use\n"
                "\t\t\"pip install pillow\""
            )

        image: Image.Image
        self._image = image

        self._binds = []

        self.style = style if style is not ... else Style()
        self.active_style = active_style if active_style\
            is not ... else Style()
        self.hover_style = hover_style if hover_style is not ... else Style()

        if min_width is not ...:
            self._width = min_width

        if min_height is not ...:
            self._height = min_height

        self.__parent = parent

        self.__parent.theme.notify_on(
            ThemeManager.NotifyEvent.theme_reload,
            self.notify
        )

        if all([
            "border_top_left_radius" in self.theme.frame,
            self.style.borderTopLeftRadius is ...
        ]):
            self.style.borderTopLeftRadius = self.theme.frame.border_top_left_radius
            if border_top_left_radius is not ...:
                self.style.borderTopLeftRadius = border_top_left_radius

        if all([
            "border_top_right_radius" in self.theme.frame,
            self.style.borderTopRightRadius is ...
        ]):
            self.style.borderTopRightRadius = self.theme.frame.border_top_right_rasdius
            if border_top_right_radius is not ...:
                self.style.borderTopRightRadius = border_top_right_radius

        if all([
            "border_bottom_left_radius" in self.theme.frame,
            self.style.borderBottomLeftRadius is ...
        ]):
            self.style.borderBottomLeftRadius = self.theme.frame.border_bottom_left_radius
            if border_bottom_left_radius is not ...:
                self.style.borderBottomLeftRadius = border_bottom_left_radius

        if all([
            "border_bottom_right_radius" in self.theme.frame,
            self.style.borderBottomRightRadius is ...
        ]):
            self.style.borderBottomRightRadius = self.theme.border_bottom_right_radius
            if border_bottom_right_radius is not ...:
                self.style.borderBottomRightRadius = border_bottom_right_radius

        # border config
        if self.style.borderWidth is ...:
            self.style.borderWidth = 0

            if "border_width" in self.theme.frame:
                self.style.borderWidth = self.theme.frame.border_width

        if margin is ...:
            margin = self.theme.frame.margin if "margin" in\
                                                self.theme.frame else 0

        if padding is ...:
            padding = self.theme.frame.padding if "padding" in\
                                                  self.theme.frame else 0

        if self.style.margin is ...:
            self.style.margin = margin

        if self.style.padding is ...:
            self.style.padding = padding

        super().__init__(layout, self.style.margin, self.style.padding)

        self.style.notify_on("margin", self.notify)
        self.style.notify_on("padding", self.notify)

        # arguments
        if width is not ... and self.style.width is ...:
            self.style.width = width

        if height is not ... and self.style.height is ...:
            self.style.height = height

        if self.style.backgroundColor is ...:
            self.style.backgroundColor = self.theme.frame.bg1 if bg is ... else bg
            if isinstance(self.__parent, Frame) and \
                    self.__parent.style.backgroundColor == self.theme.frame.bg1:
                self.style.backgroundColor = self.theme.frame.bg2 if bg is ... else bg

        if border_width is not ... and self.style.borderWidth == 0:
            self.style.borderWidth = border_width

        if self.style.borderColor is ...:
            self.style.borderColor = self.theme.frame.border\
                if border_color is ... else border_color

        # border radii
        if border_radius is ... and "border_radius" in self.theme.frame:
            border_radius = self.theme.frame.border_radius

        if border_bottom_radius is ... and\
                "border_bottom_radius" in self.theme.frame:
            border_bottom_radius = self.theme.frame.border_bottom_radius

        if border_top_radius is ... and\
                "border_top_radius" in self.theme.frame:
            border_top_radius = self.theme.frame.border_top_radius

        if self.style.borderTopLeftRadius is ...:
            self.style.borderTopLeftRadius = border_radius

        if self.style.borderTopRightRadius is ...:
            self.style.borderTopRightRadius = border_radius

        if self.style.borderBottomLeftRadius is ...:
            self.style.borderBottomLeftRadius = border_radius

        if self.style.borderBottomRightRadius is ...:
            self.style.borderBottomRightRadius = border_radius

        if border_top_radius is not ...:
            if self.style.borderTopLeftRadius is ...:
                self.style.borderTopLeftRadius = border_top_radius

            if self.style.borderTopRightRadius is ...:
                self.style.borderTopRightRadius = border_top_radius

        if border_bottom_radius is not ...:
            if self.style.borderBottomLeftRadius is ...:
                self.style.borderBottomLeftRadius = border_bottom_radius

            if self.style.borderBottomRightRadius is ...:
                self.style.borderBottomRightRadius = border_bottom_radius

    @property
    def theme(self) -> ThemeManager:
        return self.__parent.theme

    @property
    def parent(self) -> tp.Union["Frame", tp.Any]:
        return self.__parent

    @property
    def root(self) -> "PgRoot":
        return self.parent.root

    def after(
            self,
            timeout: int,
            function: tp.Callable[[tp.Any], RES_T],
            *args,
            **kwargs
    ) -> Future[RES_T]:
        """
        calls the given function after timeout milliseconds

        :param timeout: timeout in milliseconds
        :param function: the function to call
        :param args: the given functions positional arguments
        :param kwargs: the given functions keyword arguments
        :returns: a future with the function's result
        """
        return self.root.after(timeout, function, *args, **kwargs)

    @property
    def focused(self) -> bool:
        return self._focused

    def set_focus(self, **kwargs):
        """
        set this item as currently focused
        """
        self._focused = True

    def stop_focus(self):
        """
        remove focus from this item
        """
        self._focused = False

    def notify_focus(self, widget: tp.Union["GeometryManager", None] = None):
        """
        notify the root that a widget has been set as focus
        """
        self.parent.notify_focus(widget)

    def configure(self, **kwargs) -> None:
        """
        configure any of the init parameters (except parent)
        """
        for key, value in kwargs.items():
            match key:
                case "width":
                    self.width = value

                case "height":
                    self.height = value

                case "min_width":
                    self._width = value

                case "min_height":
                    self._height = value

                case "border_radius":
                    self.style.borderRadius = value

                case "border_bottom_radius":
                    self.style.borderBottomRadius = value

                case "border_top_radius":
                    self.style.borderTopRadius = value

                case _:
                    # check if in display config
                    new_key = display_configurify(key)

                    if new_key in self.style:
                        if isinstance(value, type(self.style[new_key])):
                            raise TypeError(
                                f"Can't change \"{self.style[new_key]}\" to \"{value}\": "
                                f"invalid type!"
                            )

                        self.style[key] = value

                    elif key in self.layout_params:
                        if isinstance(value, type(self.style[new_key])):
                            raise TypeError(
                                f"Can't change \"{self.layout_params[new_key]}\" to \"{value}\": "
                                f"invalid type!"
                            )

                        self.layout_params[key] = value

    # interfacing
    def notify(
            self,
            event: ThemeManager.NotifyEvent | Style.NotifyEvent,
            info: tp.Any = ...
    ) -> None:
        """
        gets called by another class
        """
        match event:
            case ThemeManager.NotifyEvent.theme_reload:
                print("theme changed")
                # the theme has been reloaded
                # if not self._display_config_configured.bg:
                #     if isinstance(self.__parent, Frame) and self.__parent._display_config["bg"] == self.theme.frame.bg1:
                #         self._display_config.bg = self.theme.frame.bg2
                #
                #     else:
                #         self._display_config.bg = self.theme.frame.bg1
                #
                # if not self._display_config_configured.border_color:
                #     self._display_config.border_color = self.theme.frame.border
                #
                # if not self._display_config_configured.border_radius:
                #     self._display_config.border_radius = self.theme.frame.border_radius

            case Style.NotifyEvent.property_change:
                print(f"style changed: {info}, new value: {self.style[info]}")

                if info in ("padding", "margin"):
                    self.layout_params[info] = self.style[info]

            case _:
                super().notify(event, info)

    def get_size(self) -> tuple[int, int]:
        """
        get the frames size (including children)
        """
        # width
        width = self._width if self._width_configured else \
            self.assigned_width

        # height
        height = self._height if self._height_configured else \
            self.assigned_height

        return width.__floor__(), height.__floor__()

    def get_position(self) -> tuple[int, int]:
        return self._x, self._y

    def draw(self, surface: pg.Surface) -> None:
        """
        draw the frame
        """
        current_style = self.style

        if self.is_hover:
            current_style = self.style.overwrite(self.hover_style)

        if self.is_active:
            current_style = self.style.overwrite(self.hover_style).overwrite(
                self.active_style
            )

        self.layout_params.padding = current_style.padding
        self.layout_params.margin = current_style.margin

        if current_style.height is not ...:
            self.height = current_style.height

        else:
            self.unset_height()

        if current_style.width is not ...:
            self.width = current_style.width

        else:
            self.unset_width()

        width, height = self.get_size()

        # check if the frame even exists
        if width <= 0 or height <= 0:
            return

        _surface = pg.Surface((width, height), pg.SRCALPHA)

        # draw the frame
        r_rect = pg.Rect((0, 0, width, height))

        pg.draw.rect(
            _surface,
            current_style.backgroundColor.irgba,
            r_rect,
            border_top_left_radius=current_style.borderTopLeftRadius,
            border_top_right_radius=current_style.borderTopRightRadius,
            border_bottom_left_radius=current_style.borderBottomLeftRadius,
            border_bottom_right_radius=current_style.borderBottomRightRadius
        )

        if self._image is not ...:
            width, height = self.get_size()
            image_width, image_height = self._image.size

            # try to reserve aspect angle
            if width <= 0 < height:
                width = int((image_width / image_height) * height)

            if height <= 0 < width:
                height = int((image_height / image_width) * width)

            # if no with or height has yet been set, clone the
            # original image's size
            if width <= 0:
                width = image_width

            if height <= 0:
                height = image_height

            # center image
            pos = (
                self.assigned_width / 2 - width / 2,
                self.assigned_height / 2 - height / 2
            )

            pos = (
                pos[0] if pos[0] >= 0 else 0,
                pos[1] if pos[1] >= 0 else 0
            )
            now_image = pi_image_to_surface(
                self._image.resize((width, height))
            )
            _surface.blit(now_image, pos)

        if current_style.borderWidth > 0:
            # print(f"drawing border, {current_style.borderColor}")
            pg.draw.rect(
                _surface,
                current_style.borderColor.irgba,
                r_rect,
                width=current_style.borderWidth,
                border_top_left_radius=current_style.borderTopLeftRadius,
                border_top_right_radius=current_style.borderTopRightRadius,
                border_bottom_left_radius=current_style.borderBottomLeftRadius,
                border_bottom_right_radius=current_style.borderBottomRightRadius
            )

        # draw children
        for child, params in self._child_params:
            child.draw(_surface)

        # draw
        surface.blit(_surface, (self._x, self._y))

    def place(
            self,
            x: int,
            y: int,
    ) -> None:
        """
        place the frame in a parent container

        :param x: x-position
        :param y: y-position
        """
        if not self.parent.check_set_layout(Layout.Absolute):
            raise TypeError(
                "can't place in a container that is"
                "not managed by \"Absolute\""
            )

        self.parent.set_layout(Layout.Absolute)

        self.__parent.add_child(self, x=x, y=y)

    def pack(
            self,
            anchor: tp.Literal["top", "bottom", "left", "right"] = TOP,
    ) -> None:
        """
        pack the frame into a parent container

        :param anchor: where to orient the frame at (direction)
        """
        if not self.parent.check_set_layout(Layout.Pack):
            raise TypeError(
                "can't pack in a container that is not managed by \"Pack\"."
                f" Configured Manager: {self.__parent.layout}"
            )

        self.parent.set_layout(Layout.Pack)

        self.__parent.add_child(self, anchor=anchor.lower())

    def grid(
            self,
            row: int,
            column: int,
            rowspan: int = 1,
            columnspan: int = 1,
            sticky: str = "",
            margin: int = 0,
    ) -> None:
        """
        grid the frame into a parent container

        :param row: the row the item should be placed in
        :param column: the column the item should be placed in
        :param rowspan: how many rows the widget should occupy
        :param columnspan: how many columns the widget should occupy
        :param sticky: expansion, can be a combination of "n", "e", "s", "w"
        :param margin: the distance to the grids borders
        """
        if not self.parent.check_set_layout(Layout.Grid):
            raise TypeError(
                "can't grid in a container that is not managed by \"Grid\""
            )

        self.parent.set_layout(Layout.Grid)

        self.__parent.add_child(
            self,
            row=row,
            column=column,
            sticky=sticky,
            margin=margin,
            rowspan=rowspan,
            columnspan=columnspan
        )

    def set_position(self, x: int, y: int) -> None:
        """
        set the child's position (used by parents)
        """
        self._x = x
        self._y = y

    def set_size(self, width: float, height: float) -> None:
        """
        set the child's size (used by parens)
        """
        self.width = width
        self.height = height

    def bind(
            self,
            sequence: FrameBind,
            func: tp.Callable[[tp.Any], tp.Any | None] = ...,
            add: tp.Literal["", "-"] | bool | None = ...
    ) -> int:
        """
        bind a callback to an event

        :param sequence: the event to listen for
        :param func: the function callback
        :param add: in case you want multiple callbacks for one event
        :returns: function id (for unbinding)
        """
        # if add isn't specified, remove all already existing binds
        if not add or add != "+":
            for b in self._binds.copy():
                if b["event"] == sequence:
                    self._binds.remove(b)

        new_id = 0
        if self._binds:
            new_id = max([e["id"] for e in self._binds]) + 1

        self._binds.append({
            "id": new_id,
            "event": sequence,
            "function": func
        })

        return new_id

    def unbind(self, func_id: int) -> None:
        """
        removes a bind event
        """
        for bind in self._binds:
            if bind["id"] == func_id:
                self._binds.remove(bind)
                return

    def _execute_event(self, event: FrameBind, *args, **kwargs) -> None:
        """
        check if a bind exists for the given event and if one does,
        execute it with the given parameters
        """
        for b in self._binds:
            if b["event"].value == event.value:
                b["function"](*args, **kwargs)

    def _on_hover(self) -> None:
        """
        called on hover
        """
        self._execute_event(FrameBind.hover)

    def _on_active(self) -> None:
        """
        called on active
        """
        self.root.set_focus(self)
        self._execute_event(FrameBind.active)

    def _on_no_active_hover(
            self,
            from_active: bool = False,
            from_hover: bool = False
    ) -> None:
        """
        called on no active / hover
        """
        if from_active:
            self._execute_event(FrameBind.active_release)

        if from_hover:
            self._execute_event(FrameBind.hover_release)

    def delete(self) -> None:
        """
        delete the widget
        """
        self.__parent = ...

        # terminate all children
        for child in self._children:
            child.delete()

# TODO: resizing issue
# TODO: image border radius



"""
_tk_vars.py
15. May 2023

implements tkinter variables like StrVar and IntVar

Author:
Nilusink
"""
from copy import copy, deepcopy
from enum import Enum
import typing as tp


_T = tp.TypeVar("_T")


class _TraceModes(Enum):
    array = "array"
    read = "read"
    write = "write"
    unset = "unset"


class Variable:
    TraceModes = _TraceModes

    _master = ...
    _value: _T = ...
    _name: str = ...
    _deepcopy: bool = False

    _traces: list[tuple[
        _TraceModes | list[_TraceModes],
        tp.Callable[[tp.Self, int, _TraceModes], None]
    ]]

    def __init__(
            self,
            master=None,
            value: _T = None,
            name: str = None,
            use_deepcopy: bool = False
    ) -> None:
        self._traces = []

        self._master = master
        self._value = value
        self._name: str = name
        self._deepcopy = use_deepcopy

    def _check_send(self, mode: _TraceModes) -> None:
        """
        checks if a trace is set and if it is, send callback
        """
        for m, c in self._traces:
            if any([
                isinstance(m, list) and mode in m,
                m == mode
            ]):
                c(self, 0, mode)

    @property
    def name(self) -> str:
        return self._name

    def get(self) -> _T:
        """
        get the variable
        """
        self._check_send(_TraceModes.read)
        return deepcopy(self._value) if self._deepcopy else copy(self._value)

    def set(self, value: _T) -> None:
        """
        set the variable
        """
        self._value = value
        self._check_send(_TraceModes.write)

    def trace_add(
            self,
            mode: _TraceModes | list[_TraceModes],
            callback_name: tp.Callable[[tp.Self, int, _TraceModes], None]
    ) -> None:
        """
        create a trace
        """
        if mode == _TraceModes.array:
            raise RuntimeWarning("mode \"array\" is currently not implemented")

        # noinspection PyTypeChecker
        self._traces.append((mode, callback_name))

    def trace_remove(
            self,
            mode: _TraceModes | list[_TraceModes],
            callback_name: tp.Callable[[tp.Self, int, _TraceModes], None]
    ) -> None:
        """
        delete a trace
        """
        for i, (m, c) in enumerate(self._traces.copy()):
            if m == mode and c == callback_name:
                self._traces.pop(i)

    def trace_info(self) -> None:
        """
        trace infos
        """
        for m, c in enumerate(self._traces):
            print(f"{m}: {c}")

    def __iadd__(self, other: _T) -> tp.Self:
        self.set(self._value + other)
        return self

    def __isub__(self, other: _T) -> tp.Self:
        self.set(self._value - other)
        return self

    def __del__(self) -> None:
        """
        class delete
        """
        self._check_send(_TraceModes.unset)


# Actual TkVariables:
class StringVar(Variable):
    def __init__(
            self,
            master=None,
            value: str = None,
            name: str = None
    ) -> None:
        super().__init__(master, value, name)

    def set(self, value: str) -> None:
        super().set(value)

    def get(self) -> str:
        return super().get()


class IntVar(Variable):
    def __init__(
            self,
            master=None,
            value: int = None,
            name: str = None
    ) -> None:
        super().__init__(master, value, name)

    def set(self, value: int) -> None:
        super().set(value)

    def get(self) -> int:
        return super().get()


class DoubleVar(Variable):
    def __init__(
            self,
            master=None,
            value: float = None,
            name: str = None
    ) -> None:
        super().__init__(master, value, name)

    def set(self, value: float) -> None:
        super().set(value)

    def get(self) -> float:
        return super().get()


class BooleanVar(Variable):
    def __init__(
            self,
            master=None,
            value: bool = None,
            name: str = None
    ) -> None:
        super().__init__(master, value, name)

    def set(self, value: bool) -> None:
        super().set(value)

    def get(self) -> bool:
        return super().get()




"""
_geo_types.py
04. February 2023

types for the geometry manager

Author:
Nilusink
"""
from enum import Enum


class Layout(Enum):
    Absolute = "absolute"
    Pack = "pack"
    Grid = "grid"


class GeoNotes(Enum):
    SetNormal = 0
    SetHover = 1
    SetActive = 2



"""
_surface_frame.py
24. May 2023

A pygame surface to draw on

Author:
Nilusink
"""
from ._frame import Frame
import typing as tp
import pygame as pg


class SurfaceFrame(Frame):
    """
    A pygame surface to draw on
    """
    _surface: pg.Surface

    def __init__(
            self,
            parent: tp.Union["Frame", tp.Any],
            width: int = ...,
            height: int = ...,
            margin: int = ...,
            min_width: int = ...,
            min_height: int = ...,
    ) -> None:

        # limit the arguments that get passed to the parent
        super().__init__(
            parent=parent,
            width=width,
            height=height,
            margin=margin,
            min_width=min_width,
            min_height=min_height
        )

    @property
    def surface(self) -> pg.Surface:
        return self._surface

    def draw(self, surface: pg.Surface) -> None:
        """
        insert the surface
        """
        surface.blit(self._surface, (self._x, self._y))



"""
_funcs.py
05. February 2023

a few generic useful functions

Author:
Nilusink
"""
import typing as tp


def arg_or_default(
        value: tp.Any,
        default_value: tp.Any,
        check_if: tp.Any = ...
) -> tp.Any:
    """
    :param value: the value to check
    :param default_value: what the value should be if it equals `check_if`
    :param check_if: what to check for
    """
    return default_value if value is check_if else value


def point_in_box(
        point: tuple[int, int],
        box: tuple[int, int, int, int]
) -> bool:
    """
    check if a point is inside a box

    :param point: point to check
    :param box: x, y width, height
    """
    return all([
        # x collision
        box[0] <= point[0],
        point[0] <= box[0] + box[2],

        # y collision
        box[1] <= point[1],
        point[1] <= box[1] + box[3]
    ])



"""
_entry.py
22. February 2023

<description>

Author:
Nilusink
"""
from ..types import (
    GeoNotes, KeyboardNotifyEvent, FrameBind, StringVar, Scheme, Style, Color,
)
from ..theme import ThemeManager
from contextlib import suppress
from ._label import Label
from ._frame import Frame
import typing as tp
import pygame as pg
import time


class Entry(Frame):
    _label: Label = ...
    _string_var: StringVar = ...
    _placeholder_text: str = ...
    _placeholder_color: Color = ...
    _cursor_pos: int = 0
    _max_length: int | None = None
    _allowed: str = ...
    _scheme: Scheme
    _f_lower: bool = False
    _f_upper: bool = False

    def __init__(
            self,
            parent: tp.Union["Frame", tp.Any],
            width: int = ...,
            height: int = ...,
            border_width: int = ...,
            border_color: Color = ...,
            textvariable: StringVar = ...,
            placeholder_text: str = ...,
            placeholder_color: Color = ...,
            bg: Color = ...,
            fg: Color = ...,
            valid_symbols: str = ...,
            scheme: Scheme = ...,
            force_uppercase: bool = False,
            force_lowercase: bool = False,
            max_length: int | None = None,
            **kwargs
    ) -> None:
        """
        A widget where you can enter text

        :param valid_symbols: if given, the entry only allows the given
        characters. If scheme is given, this option will be ignored
        """
        self._f_lower = force_lowercase
        self._f_upper = force_uppercase
        self._max_length = max_length
        self._allowed = valid_symbols
        self._scheme = scheme

        if placeholder_text is ...:
            placeholder_text = "Entry"

        if placeholder_color is ...:
            placeholder_color = Color.from_hex("#666")

        self._placeholder_text = placeholder_text
        self._placeholder_color = placeholder_color

        # initialize text-variable
        self._string_var = textvariable
        if textvariable is ...:
            self._string_var = StringVar(value="")

        # set theme colors
        if bg is ... and "bg" in parent.theme.entry:
            bg = parent.theme.entry.bg

        if fg is ... and "fg" in parent.theme.entry:
            fg = parent.theme.entry.fg

        if border_width is ... and "border_width" in parent.theme.entry:
            border_width = parent.theme.entry.border_width

        if border_color is ... and "border_color" in parent.theme.entry:
            border_color = parent.theme.entry.border_color

        hover_border_color = ...
        active_border_color = ...
        if "hover_border_color" in parent.theme.entry:
            hover_border_color = parent.theme.entry.hover_border_color

        if "active_border_color" in parent.theme.entry:
            active_border_color = parent.theme.entry.active_border_color

        if "style" in kwargs:
            if kwargs["style"].backgroundColor is ...:
                kwargs["style"].backgroundColor = bg

            if kwargs["style"].color is ...:
                kwargs["style"].color = fg

            if kwargs["style"].border_width is ...:
                kwargs["style"].border_width = border_width

            if kwargs["style"].borderColor is ...:
                kwargs["style"].borderColor = border_color

        # create hover and active style if they don't exist in kwargs
        if "hover_style" not in kwargs:
            kwargs["hover_style"] = Style()

        if "active_style" not in kwargs:
            kwargs["active_style"] = Style()

        # apply default arguments
        if kwargs["hover_style"].borderColor is ...:
            kwargs["hover_style"].borderColor = hover_border_color

        if kwargs["active_style"].borderColor is ...:
            kwargs["active_style"].borderColor = active_border_color

        super().__init__(
            parent=parent,
            width=width,
            height=height,
            border_width=border_width,
            border_color=border_color,
            bg=bg,
            min_width=100,
            min_height=10,
            **kwargs
        )
        self.style.color = fg

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self._label = Label(
            parent=self,
            text=""
        )
        self._label.grid(0, 0, sticky="nsew")

        self._sync_label_style("color", "normal")
        self._sync_label_style("color", "hover")
        self._sync_label_style("color", "active")

        self.style.notify_on(
            "color", lambda *_: self._sync_label_style("color")
        )
        self.hover_style.notify_on(
            "color", lambda *_: self._sync_label_style("color", "hover")
        )
        self.active_style.notify_on(
            "color", lambda *_: self._sync_label_style("color", "active")
        )

    def _sync_label_style(
            self, key: str,
            style_type: tp.Literal["normal", "hover", "active"] = "normal"
    ) -> None:
        """
        sync a label style with the button style
        """
        if style_type == "normal":
            self._label.style[key] = self.style[key]

        elif style_type == "hover":
            self._label.hover_style[key] = self.hover_style[key]

        elif style_type == "active":
            self._label.active_style[key] = self.active_style[key]

        else:
            raise ValueError("Invalid label sync type: ", style_type)

    # focus stuff
    def _on_hover(self) -> None:
        """
        change mouse
        """
        pg.mouse.set_cursor(pg.SYSTEM_CURSOR_IBEAM)
        super()._on_hover()

    def _on_no_active_hover(
            self,
            from_active: bool = False,
            from_hover: bool = False
    ) -> None:
        pg.mouse.set_cursor(pg.SYSTEM_CURSOR_ARROW)

        super()._on_no_active_hover(from_active, from_hover)

    def stop_focus(self) -> None:
        """
        called when the entry is unfocused
        """
        self._on_no_active_hover(True, False)
        super().stop_focus()

    def draw(self, surface: pg.Surface) -> None:
        """
        insert cursor if active
        """
        self._is_active = self.focused
        bg_width, bg_height = self._label.get_size()

        current_style = self.style

        if self.is_hover:
            current_style = self.style.overwrite(self.hover_style)

        if self.is_active:
            current_style = self.style.overwrite(self.hover_style).overwrite(
                self.active_style
            )

        # set the label color to placeholder_color if no text has been entered
        is_placeholder = not self._string_var.get()
        if is_placeholder:
            self._label.text_variable.set(self._placeholder_text)
            self._label.style.color = self._placeholder_color

        else:
            self._label.text_variable.set(self._string_var.get())
            self._label.style.color = self.style.color

        super().draw(surface)

        # cursor only when active and blink
        if self.focused and int((2 * time.time()) % 2):
            # calculate cursor position
            x, y = self._label.get_position()
            width, height = self._label.text_size

            # get frame center
            center_x = self._x + x + bg_width / 2
            center_y = self._y + y + bg_height / 2

            cursor_text = self._label._font.render(
                self._string_var.get()[:self._cursor_pos],
                True,
                current_style.color.irgba
            )
            x_off = cursor_text.get_size()[0]

            # center label on frame center
            label_end = (
                center_x - (0 if is_placeholder else width / 2) + x_off,
                center_y - height / 2
            )

            pg.draw.rect(
                surface,
                current_style.color.irgb,
                pg.Rect(label_end, (3, height))
            )

    def notify(
            self,
            event: (
                    ThemeManager.NotifyEvent
                    | Style.NotifyEvent
                    | KeyboardNotifyEvent
            ),
            info: tp.Any = ...
    ) -> None:
        match event:
            case KeyboardNotifyEvent.key_down:
                if info.key == pg.K_BACKSPACE:
                    s_len = len(self._string_var.get())

                    # deletion if a scheme is given
                    if self._scheme is not ...:
                        if s_len > 0:
                            if self._cursor_pos < s_len:
                                # split string at the cursor and replace
                                # the next character with a space character
                                prev_string = self._string_var.get()
                                self._string_var.set(
                                    prev_string[
                                        :self._cursor_pos
                                    ]
                                    + " "
                                    + prev_string[self._cursor_pos + 1:]
                                )

                            else:
                                # split the string at the cursor and delete
                                # the next character
                                prev_string = self._string_var.get()
                                self._string_var.set(
                                    prev_string[
                                        :self._cursor_pos - 1
                                    ]
                                    + prev_string[self._cursor_pos:]
                                )

                                # de-increment cursor (if not already at 0)
                                if self._cursor_pos >= 0:
                                    self._cursor_pos -= 1

                        return

                    # normal deletion
                    if s_len > 0:
                        # split the string at the cursor and delete
                        # the next character
                        prev_string = self._string_var.get()
                        self._string_var.set(
                            prev_string[:self._cursor_pos-1]
                            + prev_string[self._cursor_pos:]
                        )

                        # de-increment cursor (if not already at 0)
                        if self._cursor_pos >= 0:
                            self._cursor_pos -= 1

                elif info.key == pg.K_RETURN:
                    self.root.set_focus(self.root)
                    self._execute_event(FrameBind.key_return)

                elif info.key == pg.K_LEFT:
                    if self._cursor_pos >= 0:
                        self._cursor_pos -= 1

                elif info.key == pg.K_RIGHT:
                    if self._cursor_pos < len(self._string_var.get()):
                        self._cursor_pos += 1

                # if none of the above apply and a unicode character is
                # present, add it to the string
                elif info.unicode:
                    override = False
                    set_character = None
                    next_valid = False

                    # force capitalize - lower the character
                    if self._f_upper:
                        info.unicode = info.unicode.upper()

                    if self._f_lower:
                        info.unicode = info.unicode.lower()

                    # check max length
                    if self._max_length is not None and\
                            self._cursor_pos >= self._max_length:
                        return

                    # check valid symbols
                    if self._allowed is not ... and\
                            info.unicode not in self._allowed\
                            and self._scheme is ...:
                        return

                    if self._scheme is not ...:
                        override = True
                        try:
                            # check current character
                            is_valid, punct = self._scheme.validate(
                                self._cursor_pos,
                                info.unicode
                            )

                            # check next character
                            with suppress(IndexError):
                                next_valid, _ = self._scheme.validate(
                                    self._cursor_pos + 1, info.unicode
                                )

                            if is_valid:
                                if punct is not None:
                                    set_character = punct

                            else:
                                return

                        except IndexError:
                            return

                    # insert character at cursor position
                    prev_string = self._string_var.get()
                    self._string_var.set(
                        prev_string[:self._cursor_pos]
                        + (set_character if set_character is not None else "")
                        + info.unicode
                        + prev_string[self._cursor_pos+override:]
                    )
                    self._cursor_pos += 1 + (
                            next_valid and set_character is not None
                    )

            case GeoNotes.SetActive:
                new_index = self._label.get_index_on_position(
                    info[0] - self._x
                )

                if 0 <= new_index <= len(self._string_var.get()):
                    self._cursor_pos = new_index

                print("index: ", new_index)

                if self._notify_child_active_hover(info):
                    self.set_no_hover_active()
                else:
                    self.set_active()

            case _:
                super().notify(event, info)


# TODO: clickable cursor
# TODO: enabled / disabled
# TODO: not shown if no height | width | sticky is given



"""
_button.py
08. March 2023

Something you can click

Author:
Nilusink
"""
from ..types import Color, Layout, Style, Variable
from ..utils import arg_or_default
from ._label import Label
from ._frame import Frame
from copy import copy
import typing as tp


class Button(Frame):
    _label: Label = ...
    _command: tp.Callable[[], None] = ...
    _text_var: Variable = ...

    def __init__(
            self,
            parent: tp.Union["Frame", tp.Any],
            *,
            bg: Color = ...,
            fg: Color = ...,
            text: str = "Button",
            width: int = ...,
            height: int = ...,
            command: tp.Callable[[], None] = ...,
            textvariable: Variable = ...,
            **kwargs
    ) -> None:
        """

        """
        self._command = command
        self._text_var = textvariable

        # set theme colors
        if bg is ... and "bg" in parent.theme.button:
            bg = parent.theme.button.bg

        if fg is ... and "fg" in parent.theme.button:
            fg = parent.theme.button.fg

        hover_bg = hover_fg = ...
        active_bg = active_fg = ...
        if "hover_bg" in parent.theme.button:
            hover_bg = parent.theme.button.hover_bg

        if "hover_fg" in parent.theme.button:
            hover_fg = parent.theme.button.hover_fg

        if "active_bg" in parent.theme.button:
            active_bg = parent.theme.button.active_bg

        if "active_fg" in parent.theme.button:
            active_fg = parent.theme.button.active_fg

        if "style" in kwargs:
            if kwargs["style"].backgroundColor is ...:
                kwargs["style"].backgroundColor = bg

            if kwargs["style"].color is ...:
                kwargs["style"].color = fg

        # create hover and active style if they don't exist in kwargs
        if "hover_style" not in kwargs:
            kwargs["hover_style"] = Style()

        if "active_style" not in kwargs:
            kwargs["active_style"] = Style()

        # apply default arguments
        if kwargs["hover_style"].backgroundColor is ...:
            kwargs["hover_style"].backgroundColor = hover_bg

        if kwargs["hover_style"].color is ...:
            kwargs["hover_style"].color = hover_fg

        if kwargs["active_style"].backgroundColor is ...:
            kwargs["active_style"].backgroundColor = active_bg

        if kwargs["active_style"].color is ...:
            kwargs["active_style"].color = active_fg

        # initialize parent
        super().__init__(
            parent=parent,
            width=width,
            height=height,
            bg=bg,
            min_width=20,
            min_height=20,
            **kwargs
        )
        self.set_layout(Layout.Grid)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.style.color = arg_or_default(fg, Color.white(), ...)

        label_hover: Style = copy(kwargs["hover_style"])
        label_active: Style = copy(kwargs["active_style"])
        label_hover.backgroundColor = Color.transparent()
        label_active.backgroundColor = Color.transparent()

        self._label = Label(
            parent=self,
            text=text,
            textvariable=textvariable,
            fg=fg,
            style=Style(backgroundColor=Color.transparent()),
            hover_style=label_hover,
            active_style=label_active
        )
        self._label.grid(row=0, column=0, sticky="nsew")

        # first time style sync
        self._sync_label_style("color", "normal")

    def draw(self, surface) -> None:
        super().draw(surface)

    def _sync_label_style(
            self,
            key: str,
            style_type: tp.Literal["normal", "hover", "active"] = "normal"
    ) -> None:
        """
        sync a label style with the button style
        """
        if style_type == "normal":
            self._label.style[key] = self.style[key]

        elif style_type == "hover":
            self._label.hover_style[key] = self.hover_style[key]

        elif style_type == "active":
            self._label.active_style[key] = self.active_style[key]

        else:
            raise ValueError("Invalid label sync type: ", style_type)

    def _on_hover(self) -> None:
        """
        set the label to hover
        """
        self._label.set_hover(False)

        super()._on_hover()

    def _on_active(self) -> None:
        """
        set the label to active
        """
        self._label.set_active(False)

        # execute assigned command
        if self._command is not ...:
            self._command()

        super()._on_active()

    def _on_no_active_hover(
            self,
            from_active: bool = False,
            from_hover: bool = False
    ) -> None:
        """
        set the label to no active-hover
        """
        self._label.set_no_hover_active(False)

        super()._on_no_active_hover(from_active, from_hover)



"""
_binds.py
25. May 2023

Defines Enums for all possible bind types

Author:
Nilusink
"""
from enum import Enum


class FrameBind(Enum):
    """
    the string values represent the tkinter bind names
    """
    active_move = "<B1-Motion>"
    active_release = "<ButtonRelease-1>"
    active = "<Button-1>"

    hover = "<Enter>"
    hover_release = "<Leave>"
    hover_move = "<H-Motion>"

    mousewheel = "<MouseWheel>"

    key_return = "<Return>"




"""
_manager.py
04. February 2023

The built-in theme manager

Author:
Nilusink
"""
from enum import Enum
from ..types import *
import typing as tp
import json
import os


DEFAULT_THEME: str = os.path.dirname(__file__) + "/themes/default.json"


class Appearance(Enum):
    dark = 0
    light = 1


class _NotifyEvent(Enum):
    theme_reload = 0


class ThemeManager:
    """
    the build-in theme manager
    """
    Appearance = Appearance
    NotifyEvent = _NotifyEvent
    _appearance: Appearance = Appearance.dark
    _notify_on: dict[NotifyEvent, list[tp.Callable]] = ...
    _config: dict[str, str | dict] = ...
    _theme_path: str = ...
    instance = ...

    def __new__(cls, *args, **kwargs):
        """
        only on instance of a theme manager should ever exist
        """
        if cls.instance is ...:
            cls.instance = super().__new__(cls, *args, **kwargs)

        return cls.instance

    def __init__(self, theme_path: str = ...) -> None:
        # load the theme
        self._theme_path = theme_path
        self._notify_on = {
            _NotifyEvent.theme_reload: []
        }

        self.reload_theme()

    def notify_on(self, event: NotifyEvent, who: tp.Callable) -> None:
        """
        notify the given class on events

        :param event: the event to notify on
        :param who: the class to notify
        """
        self._notify_on[event].append(who)

    def reload_theme(self) -> None:
        """
        reload the config theme
        """
        self._config = json.load(
            open(
                DEFAULT_THEME if self.theme_path is ... else self.theme_path,
                "r"
            )
        )

        # convert each color to a color class instance
        def convert_color(color):
            if isinstance(color, str):
                # differentiate between hex and rgb values
                if color.startswith("#"):
                    return Color.from_hex(color, 255)

                elif color.startswith("rgb"):
                    return Color.from_rgb(
                        *[int(val) for val in color.lstrip("rgb").split(",")]
                    )

                else:
                    raise ValueError(
                        f"Invalid color value in theme file: \"{color}\""
                    )

            # rgb values written as tuple
            elif isinstance(color, list):
                if isinstance(color[0], float) or isinstance(color[0], int):
                    return Color.from_rgb(*color)

                elif isinstance(color[0], list) or isinstance(color[0], str):
                    return convert_color(color[self._appearance.value])

            return color

        for key in self._config.copy():
            for ckey, color in self._config[key].items():
                self._config[key][ckey] = convert_color(color)

        # notify
        for element in self._notify_on[_NotifyEvent.theme_reload]:
            element(_NotifyEvent.theme_reload)

    def set_appearance(self, appearance: Appearance) -> None:
        """
        set the appearance theme

        :param appearance: must be contained in cls.appearance
        """
        if appearance in Appearance:
            self._appearance = appearance

        else:
            raise ValueError(
                f"Expected \"Appearance\", got "
                f"\"{appearance.__class__.__name__}\""
            )

        self.reload_theme()

    @property
    def theme_path(self) -> str:
        return self._theme_path

    def __getattr__(self, item: str) -> str | BetterDict:
        """
        for better accessibility
        """
        val = self._config[item]

        if isinstance(val, dict):
            return BetterDict(val)

        elif isinstance(val, str):
            return val

        else:
            raise ValueError(f"Invalid type for key \"{item}\": {type(val)}")

    def __getitem__(self, item: str) -> str | BetterDict:
        return self.__getattr__(item)




"""
_scheme.py
26. May 2023

A scheme for text

Author:
Nilusink
"""
import typing as tp
import string


class Scheme:
    """
    a scheme for text
    """
    _current_scheme: list[str]

    def __init__(self) -> None:
        self._current_scheme = []

    @classmethod
    def from_string(cls, scheme: str) -> tp.Self:
        """
        create a scheme based on the string representation
        """
        # parse the scheme
        parsed_scheme = []
        i = 0
        while i < len(scheme):
            now = scheme[i]

            # if special, find back
            if now == "<" and scheme[i+1] == "|":
                additional = 2
                while i + additional + 1 < len(scheme):
                    if scheme[i + additional] == "|"\
                            and scheme[i + additional + 1] == ">":
                        break

                    i += 1

                else:
                    continue

                parsed_scheme.append(
                    f"<|{scheme[i+1:i+additional]}|>"
                )

                i += additional + 2
                continue

            parsed_scheme.append(now)
            i += 1

        new_instance = cls()
        new_instance._current_scheme = parsed_scheme

        return new_instance

    def get(self) -> str:
        return "".join(self._current_scheme)

    def any(self) -> tp.Self:
        self._current_scheme.append("x")
        return self

    def lowercase(self) -> tp.Self:
        """
        any ascii lowercase letters (abcdefghijklmnopqrstuvwxyz)
        """
        self._current_scheme.append("l")
        return self

    def l(self) -> tp.Self:
        """
        same as lowercase
        """
        return self.lowercase()

    def uppercase(self) -> tp.Self:
        """
        all ascii uppercase letters (ABCDEFGHIJKLMNOPQRSTUVWXYZ)
        """
        self._current_scheme.append("u")
        return self

    def u(self) -> tp.Self:
        """
        same as uppercase
        """
        return self.uppercase()

    def letters(self) -> tp.Self:
        """
        all ascii lowercase + uppercase letters
        (abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ)
        """
        self._current_scheme.append("a")
        return self

    def a(self) -> tp.Self:
        """
        same as letters
        """
        return self.letters()

    def digits(self) -> tp.Self:
        """
        all numbers from 0 to 10 (0123456789)
        """
        self._current_scheme.append("d")
        return self

    def d(self) -> tp.Self:
        """
        same as numbers
        """
        return self.digits()

    def hexdigits(self) -> tp.Self:
        """
        all hexadecimal digits (0123456789abcdefABCDEF)
        """
        self._current_scheme.append("h")
        return self

    def h(self) -> tp.Self:
        """
        same as hexdigits
        """
        return self.hexdigits()

    def octdigits(self) -> tp.Self:
        """
        all octal digits (01234567)
        """
        self._current_scheme.append("o")
        return self

    def o(self) -> tp.Self:
        """
        same as octdigits
        """
        return self.octdigits()

    def symbols(self) -> tp.Self:
        r"""
        all symbols (!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~)
        """
        self._current_scheme.append("s")
        return self

    def s(self) -> tp.Self:
        """
        same as symbols
        """
        return self.symbols()

    def printable(self) -> tp.Self:
        r"""
        all ascii printable characters
        (
        0123456789
        abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
        !"#$%&'()*+,-./:;<=>?@[\]^_`{|}~
         \t\n\r\v\f
         )
        """
        self._current_scheme.append("p")
        return self

    def p(self) -> tp.Self:
        """
        same as printable
        """
        return self.printable()

    def specific(self, character: str) -> tp.Self:
        """
        a specific string to be inserted in between
        """
        self._current_scheme.append(f"<|{character}|>")
        return self

    def validate(self, position: int, to_check: str) -> tuple[int, str | None]:
        """
        checks if the given character fits into the desired position
        """
        to_match = self._current_scheme[position]

        if to_match.startswith("<|") and to_match.endswith("|>"):
            return 2, to_match.lstrip("<|").rstrip("|>")

        match to_match:
            case "l":
                return int(to_check in string.ascii_lowercase), None

            case "u":
                return int(to_check in string.ascii_uppercase), None

            case "a":
                return int(to_check in string.ascii_letters), None

            case "d":
                return int(to_check in string.digits), None

            case "o":
                return int(to_check in string.octdigits), None

            case "h":
                return int(to_check in string.hexdigits), None

            case "s":
                return int(to_check in string.punctuation), None

            case "p":
                return int(to_check in string.printable), None

            case _:
                raise ValueError(f"Invalid scheme: {self.get()}")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.get()}>"



"""
_supports_children.py
04. February 2023

All widgets which support parenting should inherit this

Author:
Nilusink
"""
import typing as tp


class SupportsChildren:
    _children: list[tp.Any] = ...

    def __init__(self) -> None:
        self._children = []

    def add_child(self, child: tp.Any) -> None:
        """
        add a child to the collection
        """
        self._children.append(child)



"""
_geo_manager.py
04. February 2023

How children are placed

Author:
Nilusink
"""
from ..types import Layout, BetterDict, TOP, BOTTOM, LEFT, RIGHT, GeoNotes
from ._supports_children import SupportsChildren
from ..utils import point_in_box
from copy import deepcopy
import typing as tp
import pygame as pg


if tp.TYPE_CHECKING:
    from ._frame import Frame


class GeometryManager(SupportsChildren):
    """
    Manages how children are placed inside a parent container
    """
    _layout: Layout = Layout.Absolute
    _width: float = 0   # -1 means not configured -> takes minimum size
    # required by its children
    _height: float = 0  # or the size it gets by the parents geometry manager
    assigned_width: float = 0
    assigned_height: float = 0
    _width_configured: bool = False
    _height_configured: bool = False
    _child_params: list[tuple[tp.Any, BetterDict]] = ...
    layout_params: BetterDict = ...
    _grid_params: BetterDict = ...

    _child_override: bool = False
    _is_active: bool = False
    _is_hover: bool = False

    def __init__(
            self,
            layout: Layout = ...,
            margin: int = 0,
            padding: int = 0
    ):
        super().__init__()

        if layout is not ...:
            if layout not in (Layout.Grid, Layout.Pack, Layout.Absolute):
                raise ValueError(f"Invalid layout type: {layout}")

        self._layout = Layout.Absolute if layout is ... else layout
        self.layout_params = BetterDict({
            "margin": margin,
            "padding": padding
        })
        self._child_params = []
        self._grid_params = BetterDict({
            "rows": {},
            "columns": {}
        })

    @property
    def layout(self) -> Layout:
        """
        the containers layout type
        """
        return self._layout

    @property
    def width(self) -> float:
        return self._width
    
    @width.setter
    def width(self, value: float) -> None:
        self._width = value
        self._width_configured = True

    def unset_width(self) -> None:
        """
        act as if the width has never been set
        """
        self._width_configured = False

    @property
    def height(self) -> float:
        return self._height

    @height.setter
    def height(self, value: float) -> None:
        self._height = value
        self._height_configured = True

    def unset_height(self) -> None:
        """
        act as if the height has never been set
        """
        self._height_configured = False

    @property
    def is_hover(self) -> bool:
        return self._is_hover

    @property
    def is_active(self) -> bool:
        return self._is_active

    def set_focus(self, widget: tp.Union["GeometryManager", None]):
        """
        set this item as currently focused
        """

    def stop_focus(self):
        """
        remove focus from this item
        """

    def _on_active(self) -> None:
        """
        called when button is clicked
        not implemented by default
        """

    def _on_hover(self) -> None:
        """
        called on hover
        not implemented by default
        """

    def _on_no_active_hover(
            self,
            from_active: bool = False,
            from_hover: bool = False
    ) -> None:
        """
        called when either hover or active goes back to normal
        not implemented by default
        """

    def set_active(self, override: bool = False) -> None:
        """
        change the widget to an active state
        """
        if not self._child_override:
            if not self._is_active:  # only call on new event
                self._on_active()

            self._is_hover = False
            self._is_active = True

        self._child_override = self._child_override or override

    def set_hover(self, override: bool = False) -> None:
        """
        set the widget to a hover state
        """
        if not self._child_override:
            if not self._is_hover:  # only call on new event
                self._on_hover()

            self._is_hover = True
            self._is_active = False

        self._child_override = self._child_override or override

    def set_no_hover_active(self, override: bool = False) -> None:
        """
        set the widget to "normal" mode
        """
        if not self._child_override:
            if self._is_hover or self._is_active:  # only call no new event
                self._on_no_active_hover(self._is_active, self._is_hover)

            self._is_hover = False
            self._is_active = False

        self._child_override = self._child_override or override

    def _notify_child_active_hover(self, mouse_pos: tuple[int, int]) -> bool:
        has_hit: bool = False

        for child in self._children:
            child: Frame

            pos = child.get_position()
            size = child.get_size()

            # noinspection PyTypeChecker
            if point_in_box(mouse_pos, (*pos, *size)):
                has_hit = True
                child.notify(
                    GeoNotes.SetActive if pg.mouse.get_pressed(3)[0]
                    else GeoNotes.SetHover,
                    (mouse_pos[0] - pos[0], mouse_pos[1] - pos[1])
                )
                continue

            child.notify(GeoNotes.SetNormal)

        return has_hit

    def notify(self, event, info) -> None:
        """
        for notifications from child / parent classes
        """
        self._child_override = False
        match event:
            case GeoNotes.SetHover:
                if self._notify_child_active_hover(info):
                    self.set_no_hover_active()
                else:
                    self.set_hover()

            case GeoNotes.SetActive:
                if self._notify_child_active_hover(info):
                    self.set_no_hover_active()
                else:
                    self.set_active()

            case GeoNotes.SetNormal:
                # also set children to inactive
                self._notify_child_active_hover((-1, -1))

                self.set_no_hover_active()

    # layout configuration
    def set_layout(self, layout: Layout) -> None:
        """
        set the container's layout type
        """
        if layout not in Layout:
            raise ValueError("Invalid container layout: ", layout)

        if self._layout != layout:
            if len(self._child_params) > 0:
                # if children are already present, delete them
                for child, _param in self._child_params:
                    child.delete()

                self._child_params.clear()

                raise RuntimeWarning(
                    "changing layout with children already present!"
                )

            self._layout = layout

    def check_set_layout(self, layout: Layout) -> bool:
        """
        checks if the layout can be changed
        """
        # not a real layout type
        if layout not in Layout:
            return False

        # is already the same layout
        if self._layout == layout:
            return True

        # no children present - can be easily changed
        if len(self._children) == 0:
            return True

        return False

    def grid_columnconfigure(
            self,
            column: int | tp.Iterable[int],
            weight: float = 1
    ) -> None:
        """
        configure one or more grid columns
        """
        # if the column is given as a list,
        # call this function with an integer as parameter
        if isinstance(column, tp.Iterable):
            for c in column:
                self.grid_columnconfigure(c, weight=weight)

            return

        # check if the layout is "grid"
        if not self.check_set_layout(Layout.Grid):
            raise ValueError("Invalid geometry type for grid-configuration")

        self.set_layout(Layout.Grid)

        self._grid_params["columns"][column] = {
            "weight": weight
        }

    def grid_rowconfigure(
            self,
            row: int | tp.Iterable[int],
            weight: float = 1
    ) -> None:
        """
        configure one or more grid rows
        """
        # if the row is given a list, call this
        # function with an integer as parameter
        if isinstance(row, tp.Iterable):
            for r in row:
                self.grid_rowconfigure(r, weight=weight)

            return

        # check if the layout is "grid"
        if not self.check_set_layout(Layout.Grid):
            raise ValueError("Invalid geometry type for grid-configuration")

        self.set_layout(Layout.Grid)

        self._grid_params["rows"][row] = {
            "weight": weight
        }

    # other stuff
    def add_child(self, child: tp.Any, **params) -> None:
        """
        add a child to the collection
        """
        if child not in self._children:
            super().add_child(child)
            self._child_params.append((child, BetterDict(params)))

    def calculate_geometry(self):
        """
        calculate how each individual child should be placed
        """
        match self._layout:
            case Layout.Absolute:
                # since the positioning is absolute,
                # the children should not influence the parents size
                for child, params in self._child_params:
                    child.set_position(params.x, params.y)

            case Layout.Pack:
                directional_dict: dict[str, int | list] = {
                    "total_x": 0,
                    "total_y": 0,
                    "children": [],
                    "sizes": []
                }
                top = deepcopy(directional_dict)
                bottom = deepcopy(directional_dict)
                left = deepcopy(directional_dict)
                right = deepcopy(directional_dict)

                # get all sizes and group by anchor
                for child, param in self._child_params:
                    child_size = child.calculate_size()

                    if param.anchor == TOP:
                        top["children"].append(child)
                        top["sizes"].append(child_size)
                        top["total_x"] += child_size[0]
                        top["total_y"] += child_size[1]

                    elif param.anchor == BOTTOM:
                        bottom["children"].append(child)
                        bottom["sizes"].append(child_size)
                        bottom["total_x"] += child_size[0]
                        bottom["total_y"] += child_size[1]

                    elif param.anchor == LEFT:
                        left["children"].append(child)
                        left["sizes"].append(child_size)
                        left["total_x"] += child_size[0]
                        left["total_y"] += child_size[1]

                    elif param.anchor == RIGHT:
                        right["children"].append(child)
                        right["sizes"].append(child_size)
                        right["total_x"] += child_size[0]
                        right["total_y"] += child_size[1]

                top["total_y"] += self.layout_params.padding * len(
                    top["children"]
                ) - 1
                bottom["total_y"] += self.layout_params.padding * len(
                    bottom["children"]
                ) - 1

                left["total_x"] += self.layout_params.padding * len(
                    left["children"]
                ) - 1
                right["total_x"] += self.layout_params.padding * len(
                    right["children"]
                ) - 1

                # if not configured, set own size
                total_x = max([top["total_x"], bottom["total_x"],
                               left["total_x"] + right["total_x"]])
                total_y = max([left["total_y"], right["total_y"],
                               top["total_y"] + bottom["total_y"]])

                # add margin
                total_x += self.layout_params.margin * 2
                total_y += self.layout_params.margin * 2

                if not self._width_configured:
                    self._width = total_x

                else:
                    total_x = self._width
    
                if not self._height_configured:
                    self._height = total_y

                else:
                    total_y = self.height

                # tell the children where they should be
                y_cen = total_y / 2
                x_cen = total_x / 2

                # left
                x_now = self.layout_params.margin
                for child, size in zip(left["children"], left["sizes"]):
                    child.set_position(x_now, y_cen - size[1] / 2)
                    x_now += size[0] + self.layout_params.padding

                # right
                x_now = total_x - self.layout_params.margin
                for child, size in zip(right["children"], right["sizes"]):
                    child.set_position(x_now - size[0], y_cen - size[1] / 2)
                    x_now -= size[0] + self.layout_params.padding

                # top
                y_now = self.layout_params.margin
                for child, size in zip(top["children"], top["sizes"]):
                    child.set_position(x_cen - size[0] / 2, y_now)
                    y_now += size[1] + self.layout_params.padding

                # bottom
                y_now = total_y - self.layout_params.margin
                for child, size in zip(bottom["children"], bottom["sizes"]):
                    child.set_position(x_cen - size[0] / 2, y_now - size[1])
                    y_now -= size[1] + self.layout_params.padding

            case Layout.Grid:
                rows: list[dict[
                    str,
                    list[tuple[tp.Any, BetterDict, bool]] | float
                ]] = []
                columns: list[dict[
                    str,
                    list[tuple[tp.Any, BetterDict, bool]] | float
                ]] = []

                for child, params in self._child_params:
                    row, column = params["row"], params["column"]

                    # if row was not yet made, make all previous ones
                    min_rows = row + params.rowspan - 1
                    min_columns = column + params.columnspan - 1

                    if len(rows) <= min_rows:
                        for n_row in range(len(rows), min_rows+1):
                            out = {
                                "weight": 0,
                                "children": []
                            }

                            if n_row in self._grid_params.rows:
                                config = self._grid_params.rows[n_row]

                                if "weight" in config:
                                    out["weight"] = config["weight"]

                            rows.append(out)

                    if len(columns) <= min_columns:
                        for n_col in range(len(columns), min_columns+1):
                            out = {
                                "weight": 0,
                                "children": []
                            }

                            if n_col in self._grid_params.columns:
                                config = self._grid_params.columns[n_col]

                                if "weight" in config:
                                    out["weight"] = config["weight"]

                            columns.append(out)

                    rows[row]["children"].append((child, params, True))
                    
                    # if rowspan is given, also add the widget to all the
                    # corresponding other rows
                    if "rowspan" in params:
                        for row_incrementor in range(1, params.rowspan):
                            rows[row + row_incrementor]["children"]\
                                .append(
                                (child, params, False)
                            )

                    columns[column]["children"].append((child, params, True))

                    if "columnspan" in params:
                        for column_incrementor in range(1, params.columnspan):
                            columns[column + column_incrementor]["children"]\
                                .append(
                                (child, params, False)
                            )

                matrix: list[list] = []

                # append all the children to a 2d matrix and check if one
                # cell has multiple children
                for r in range(len(rows)):
                    matrix.append([])

                    for c in range(len(columns)):
                        child = [
                            chi for chi in rows[r]["children"]
                            if chi in columns[c]["children"]
                        ]
                        child = list(child)

                        if len(child) > 1:
                            raise ValueError(
                                f"{len(child)} children assigned "
                                f"to row {r} column {c}!"
                            )

                        if child:
                            matrix[r].append(child[0])

                        else:
                            matrix[r].append(...)

                # calculate the minimal size for each row
                for r, row in enumerate(rows):
                    rows[r]["max_size"] = 0

                    for child, params, _is_widget in row["children"]:
                        child.calculate_size()
                        _, y = child.get_size()

                        y += 2 * params.margin

                        # split size between rows
                        y /= params.rowspan

                        if y > rows[r]["max_size"]:
                            rows[r]["max_size"] = y

                # calculate the minimal size for each column
                for c, column in enumerate(columns):
                    columns[c]["max_size"] = 0

                    for child, params, _is_widget in column["children"]:
                        child.calculate_size()
                        x, _ = child.get_size()

                        x += 2 * params.margin

                        # split size between columns
                        x /= params.columnspan

                        if x > columns[c]["max_size"]:
                            columns[c]["max_size"] = x

                # calculate the container size
                width, height = self.assigned_width, self.assigned_height

                # only subtract rows that don't have a weight
                min_width = sum([c["max_size"] for c in columns])
                min_height = sum([r["max_size"] for r in rows])

                if len(columns) + len(rows) > 0:
                    if width == 0 or height == 0 and not\
                            self._width_configured:
                        self._width = min_width

                    if width == 0 or height == 0 and not\
                            self._height_configured:
                        self._height = min_height

                if self._width_configured:
                    width = self._width

                if self._height_configured:
                    height = self.height

                # assign extra space
                extra_width = width - sum(
                    [c["max_size"] for c in columns if c["weight"] == 0]
                )
                extra_height = height - sum(
                    [r["max_size"] for r in rows if r["weight"] == 0]
                )

                total_row_weight = sum(
                    [row["weight"] for row in rows]
                )
                total_column_weight = sum(
                    [column["weight"] for column in columns]
                )

                # assign each row and column a specific size
                for r in range(len(rows)):
                    if total_row_weight == 0:
                        rows[r]["height"] = 0

                    else:
                        # assign either the minimum size or the
                        # calculated dynamic one
                        w_size = (
                                (rows[r]["weight"] / total_row_weight)
                                * extra_height
                        ).__floor__()
                        rows[r]["height"] = w_size

                    rows[r]["y_start"] = self.layout_params.padding / 2 + sum(
                        [prev_row["height"] for prev_row in rows[:r]]
                    )

                    for c in range(len(columns)):
                        if total_column_weight == 0:
                            columns[c]["width"] = 0

                        else:
                            # assign either the minimum size or the
                            # calculated dynamic one
                            w_size = (
                                    (
                                            columns[c]["weight"]
                                            / total_column_weight
                                    )
                                    * extra_width
                            ).__floor__()
                            columns[c]["width"] = w_size

                        columns[c]["x_start"] = self.layout_params.padding / 2\
                            + sum(
                            [prev_col["width"] for prev_col in columns[:c]]
                        )

                # place children
                for child, params in self._child_params:
                    # place the child proportional to the table and stickiness
                    # size = list(child.calculate_size())
                    size = [child._width, child._height]

                    row, column = params.row, params.column
                    sticky = params.sticky
                    margin = params.margin

                    width = sum([
                        columns[c]["width"] for c in
                        range(column, column + params.columnspan)
                    ])
                    height = sum([
                        rows[r]["height"] for r in
                        range(row, row + params.rowspan)
                    ])

                    x = columns[column]["x_start"]
                    y = rows[row]["y_start"]

                    x_diff = width - size[0]
                    y_diff = height - size[1]

                    # assign stickiness
                    if not child._width_configured:
                        if "w" in sticky:
                            size[0] += (x_diff / 2) - params.margin

                        if "e" in sticky:
                            size[0] += (x_diff / 2) - params.margin

                        child.assigned_width = size[0]

                    if not child._height_configured:
                        if "n" in sticky:
                            size[1] += (y_diff / 2) - params.margin

                        if "s" in sticky:
                            size[1] += (y_diff / 2) - params.margin

                        child.assigned_height = size[1]

                    child.set_position(x + margin, y + margin)

            case _:
                raise ValueError(f"Invalid geometry type: {self._layout}")

    def calculate_size(self) -> tuple[int, int]:
        """
        calculate how big the container should be
        """
        # make sure the geometry is up-to-date
        self.calculate_geometry()

        width = self._width
        height = self._height

        width = round(width)
        height = round(height)

        return width, height



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pginter",
    version="0.2.0",
    author="Nilusink",
    author_email="nilusink@protonmail.com",
    description="A python GUI interface, based on pygame",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Nilusink/PgInter",
    # project_urls={
    #     "Official Website": "http://info.fridrich.xyz",
    # },
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.11",
    include_package_data=True
    # data_files=[
    #     ('pginter/theme/themes/', ['src/pginter/theme/themes/default.json'])
    # ]
)



"""
_label.py
10. February 2023

A thing to display text

Author:
Nilusink
"""
from ..types import Color, Layout, Style, Variable, StringVar, GeoNotes
from ..utils import arg_or_default
from ..theme import ThemeManager
from ._frame import Frame
import pygame as pg
import typing as tp


class Label(Frame):
    """
    A thing to display text
    """
    _text_var: Variable = ...
    _text_size: tuple[int, int] = ...

    def __init__(
            self,
            parent: tp.Union["Frame", tp.Any],
            text: str = "Label",
            font_size: int = ...,
            bg: Color = ...,
            fg: Color = ...,
            textvariable: Variable = ...,
            **args
    ) -> None:
        """

        """
        self._text_size = (0, 0)
        self._text_var = arg_or_default(
            textvariable,
            StringVar(value=arg_or_default(text, "Button", ...)),
            ...
        )

        # remove configured arguments
        if "layout" in args:
            args.pop("layout")

        style = Style(
            color=arg_or_default(fg, parent.theme.label.fg),
            backgroundColor=arg_or_default(bg, parent.theme.label.bg),
            fontSize=arg_or_default(font_size, parent.theme.label.font_size)
        )

        # if style option exists, overwrite the default one
        if "style" in args:
            style = style.overwrite(args["style"])

            args.pop("style")

        # initialize parent class
        super().__init__(
            parent=parent,
            layout=Layout.Grid,
            style=style,
            **args
        )

        # initialize font
        # noinspection PyTypeChecker
        self._font = pg.font.SysFont(None, self.style.fontSize)

    @property
    def text_size(self) -> tuple[int, int]:
        return self._text_size

    @property
    def text_variable(self) -> Variable:
        return self._text_var

    def draw(self, surface: pg.Surface) -> None:
        """
        draw the Label
        """
        current_style = self.style

        if self.is_hover:
            current_style = self.style.overwrite(self.hover_style)

        if self.is_active:
            current_style = self.style.overwrite(self.hover_style).overwrite(
                self.active_style
            )

        # get text
        r_text = self._font.render(
            self._text_var.get(),
            True,
            current_style.color.irgba
        )

        width, height = r_text.get_size()
        self._text_size = width, height

        bg_width, bg_height = self.get_size()

        border_radius = max([
            current_style.borderTopLeftRadius,
            current_style.borderTopRightRadius,
            current_style.borderBottomLeftRadius,
            current_style.borderBottomRightRadius
        ])
        frame_size = (
            width + border_radius,
            height + border_radius
        )

        self.configure(min_width=frame_size[0], min_height=frame_size[1])
        self.assigned_width = frame_size[0]
        self.assigned_height = frame_size[1]

        # draw frame
        super().draw(surface)

        # get frame center
        center_x = self._x + bg_width / 2
        center_y = self._y + bg_height / 2

        # center label on frame center
        pos = (
            center_x - width / 2,
            center_y - height / 2
        )

        surface.blit(r_text, pos)

    def configure(
            self,
            text: str = ...,
            textvariable: Variable = ...,
            **kwargs
    ) -> None:
        """
        configure the label properties
        """
        if text is not ...:
            self._text_var.set(text)

        if textvariable is not ...:
            self._text_var = textvariable

    def notify(
            self, event: ThemeManager.NotifyEvent | Style.NotifyEvent,
            info: tp.Any = ...
    ) -> None:
        """
        for notifications from child / parent classes
        """
        match event:
            case GeoNotes.SetHover:
                self.parent.set_hover(True)

            case GeoNotes.SetActive:
                self.parent.set_active(True)

    def get_index_on_position(self, x_position: int) -> int:
        """
        calculates the character clicked (with x position)
        """
        text = self._text_var.get()
        whole_text = self._font.render(text,  True, (0, 0, 0, 0))
        whole_width = whole_text.get_width()

        # estimate the position by pretending every character has the
        # same width
        if whole_width != 0:
            i_estimated = int(round((x_position / whole_width) * len(text), 0))

        else:
            i_estimated = 0

        # check the exact position
        r_text = self._font.render(text[:i_estimated],  True, (0, 0, 0, 0))
        exact_x = r_text.get_size()[0]
        off = exact_x - x_position

        if off > 0:
            last = off

            while True:
                i_estimated += 1
                r_text = self._font.render(
                    text[:i_estimated], True, (0, 0, 0, 0)
                    )
                exact_x = r_text.get_size()[0]
                new = exact_x - x_position

                if new < last:
                    i_estimated -= 1
                    break

                if i_estimated > len(text):
                    break

                last = new

        elif off < 0:
            last = off

            while True:
                i_estimated += 1
                r_text = self._font.render(
                    text[:i_estimated], True, (0, 0, 0, 0)
                )
                exact_x = r_text.get_size()[0]
                new = exact_x - x_position

                if new < last:
                    i_estimated -= 1
                    break

                if i_estimated == 0:
                    break

                last = new

        print(f"offset: {off=}")

        return i_estimated
